
#include <stdio.h>
#include <string.h>   //strlen
#include <stdlib.h>   //strlen
#include <sys/socket.h>
#include <arpa/inet.h>   //inet_addr
#include <unistd.h>   //write
//#include <netinet/in.h> not sure this is needed
#include <pthread.h> //for threading , link with lpthread

//$ gcc program.c -lpthread -o NetDanceServer

int PortNumber;

#define OverallParamCount 6

typedef struct wordlistelement
{
   char* word;
   struct wordlistelement *next;
} WordListElement;

void getFractions(char *pc, char *dst)
{
   char fraction[10];
   char *pfraction;
   *dst = '\0';
   while(1)
   {
      pfraction = strchr(pc, '/');
      if(pfraction==NULL)
         return;
      if(*dst!='\0')
         strcat(dst, " ");
      pfraction--;
      while(*pfraction>='0'&&*pfraction<='9')
         pfraction--;
      pfraction++;
      sscanf(pfraction, "%s", fraction);
      pc=pfraction+strlen(fraction);
      strcat(dst, fraction);
   }
}

WordListElement* splitL(char* phrase, int wordcnt, int *wordCntFound)
{
   WordListElement 
      *WordListHead = NULL, 
      *WordListTail = NULL, 
      *NewWord = NULL;
   int len = 0;
   char *blanks = " \t\n\r";
   char *pc = phrase;
   char *wordread = (char*)malloc(strlen(phrase)+1);
   if(wordread==NULL)
   {
      printf("E001: Memory allocation failed\n");
      return NULL;
   }
   int proceed = 1;
   while (proceed&&*pc!='\0')
   {
      while(*pc!='\0'&&strchr(blanks, *pc)!=NULL)
         pc++;
      if(*pc=='\0')
         break;
      if(len<wordcnt)
         sscanf(pc, "%s", wordread);
      else
      {
         getFractions(pc, wordread);
         proceed = 0;
      }
      pc+=strlen(wordread);
      NewWord = (WordListElement*)malloc(sizeof(WordListElement));
      if(NewWord==NULL)
      {
         printf("E002: Memory allocation failed\n");
         return NULL;
      }
      NewWord->word = (char*)malloc(sizeof(strlen(wordread)+1));
      if(NewWord->word==NULL)
      {
         free(NewWord);
         printf("E003: Memory allocation failed\n");
         return NULL;
      }
      strcpy(NewWord->word, wordread);
      NewWord->next = NULL;
      if(WordListTail!=NULL)
         WordListTail->next = NewWord;
      WordListTail = NewWord;
      if(WordListHead == NULL)
         WordListHead = NewWord;
      len++;
   }
   free(wordread);
   *wordCntFound = len;
   return WordListHead;
}

WordListElement* split(char* phrase, int *wordcnt)
{
   WordListElement 
      *WordListHead = NULL, 
      *WordListTail = NULL, 
      *NewWord = NULL;
   int len = 0;
   char *blanks = " \t\n\r";
   char *pc = phrase;
   char *wordread = (char*)malloc(strlen(phrase)+1);
   if(wordread==NULL)
   {
      printf("E004: Memory allocation failed\n");
      return NULL;
   }
   while (*pc!='\0')
   {
      while(*pc!='\0'&&strchr(blanks, *pc)!=NULL)
         pc++;
      if(*pc=='\0')
         break;
      sscanf(pc, "%s", wordread);
      pc+=strlen(wordread);
      NewWord = (WordListElement*)malloc(sizeof(WordListElement));
      if(NewWord==NULL)
      {
         printf("E005: Memory allocation failed\n");
         return NULL;
      }
      NewWord->word = (char*)malloc(sizeof(strlen(wordread)+1));
      if(NewWord->word==NULL)
      {
         free(NewWord);
         printf("E006: Memory allocation failed\n");
         return NULL;
      }
      strcpy(NewWord->word, wordread);
      NewWord->next = NULL;
      if(WordListTail!=NULL)
         WordListTail->next = NewWord;
      WordListTail = NewWord;
      if(WordListHead == NULL)
         WordListHead = NewWord;
      len++;
   }
   *wordcnt = len;
   return WordListHead;
}

void WordListCleanup(WordListElement *head, int deep)
{
   WordListElement *kill = head;
   while(kill!=NULL)
   {
      if(deep)
         free(head->word);
      head = head->next;
      free(kill);
      kill = head;
   }
}

WordListElement *wordlists[10];

void WordListArrayCleanup( int len)
{
   int i;
   for(i=0;i<len;i++)
      WordListCleanup(wordlists[i], 1);
}

typedef struct clientlistelement
{
   int index;
   char *Name;
   int NewState;
   int Socket;
   char *Melody;
   char *Input0;
   struct clientlistelement *next;
} ClientListElement;

ClientListElement *ClientList = NULL, *ClientListTail = NULL;
int clientCount = 0;
pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
char **OverallParamValues = NULL;
char *OverallParams[] = {"MainPitch", "Duration", "Length", "DecayRate", "ScaleType", "PolyphonyType" };
char *Melody = NULL;
char *SharedContent = NULL;
int listening = 1;
int l;
int IdentifierCounter = 0;

void CleanupOverallParamValues()
{
   if(OverallParamValues!=NULL)
   {
      int i;
      for(i=0;i<OverallParamCount;i++)
         free(OverallParamValues[i]);
      free(OverallParamValues);
   }
}

int addClient(int i, int socket)
{
   ClientListElement* newClient;
   pthread_mutex_lock(&mutex);
   newClient = (ClientListElement*)malloc(sizeof(ClientListElement));
   if(newClient==NULL)
   {
      pthread_mutex_unlock(&mutex);
      printf("E007: Memory allocation failed\n");
      return 1;
   }
   newClient->index = i;
   newClient->Socket = socket;
   newClient->next = NULL;
   newClient->Name = (char*)malloc(8);
   if(newClient->Name==NULL)
   {
      pthread_mutex_unlock(&mutex);
      printf("E008: Memory allocation failed\n");
      free(newClient);
      return 1;
   }
   strcpy(newClient->Name, "noname");
   newClient->Melody = (char*)malloc(8);
   if(newClient->Melody==NULL)
   {
      pthread_mutex_unlock(&mutex);
      printf("E009: Memory allocation failed\n");
      free(newClient->Name);
      free(newClient);
      return 1;
   }
   strcpy(newClient->Melody, "0 0 0 0");
   newClient->NewState = 1;
   newClient->Input0 = (char*)malloc(8);
   if(newClient->Input0==NULL)
   {
      pthread_mutex_unlock(&mutex);
      free(newClient->Name);
      free(newClient->Melody);
      free(newClient);
      printf("E010: Memory allocation failed\n");
      return 1;
   }
   strcpy(newClient->Input0, "0 0 0 0");
   if(ClientListTail!=NULL)
      ClientListTail->next = newClient;
   ClientListTail = newClient;
   if(ClientList==NULL)
      ClientList = newClient;
   clientCount++;
   pthread_mutex_unlock(&mutex);
   return 0;
}

void removeClient(int i)
{
   pthread_mutex_lock(&mutex);
   ClientListElement *currClient = ClientList;
   ClientListElement *prev = ClientList;
   while(currClient!=NULL)
   {
      if(currClient->index == i)
      {
         if(prev==currClient)
            ClientList = ClientList->next;
         else
            prev->next = currClient->next;
         if(currClient->Name!=NULL)
            free(currClient->Name);
         if(currClient->Melody!=NULL)
            free(currClient->Melody);
         if(currClient->Input0!=NULL)
            free(currClient->Input0);
         free(currClient);
         break;
      }
      prev = currClient;
      currClient = currClient->next;
   }
   clientCount--;
   printf("There are %d clients left.", clientCount);
   pthread_mutex_unlock(&mutex);
}


int UpdateParams(char* p)
{
   int i;
   int len;
   int res = 0;
   WordListElement *params;
   WordListElement *nextparam, *oldparam;
   pthread_mutex_lock(&mutex);
   params = splitL(p, OverallParamCount-1, &len);
   nextparam = params;
   if (len<OverallParamCount)
      res = -1;
   else
   {
      CleanupOverallParamValues();
      OverallParamValues = (char**)malloc(len*sizeof(char*));
      if(OverallParamValues==NULL)
      {
         printf("E011: Memory allocation failed\n");
         return -1;
      }
      for(i=0;i<OverallParamCount-1;i++)
      {
         OverallParamValues[i] = nextparam->word;
         oldparam = nextparam;
         nextparam = nextparam->next;
         free(oldparam);
      }
      OverallParamValues[OverallParamCount-1]=nextparam->word;
      free(nextparam);
   }
   pthread_mutex_unlock(&mutex);
   return res;
}

int UpdateSharedContent()
{
   int length;
   ClientListElement* currClient = ClientList;
   pthread_mutex_lock(&mutex);
   length += (strlen("more lines\n")+
      strlen(OverallParamValues[0])+1+
      strlen(OverallParamValues[1])+1+
      strlen(OverallParamValues[2])+1+
      strlen(OverallParamValues[3])+1+
      strlen(OverallParamValues[4])+1+
      strlen(OverallParamValues[5])+1);
   length += (strlen(Melody)+1);
   if(SharedContent!=NULL)
      free(SharedContent);
   SharedContent = (char*)malloc(length);
   if(SharedContent==NULL)
   {
      printf("E012: Memory allocation failed\n");
      pthread_mutex_unlock(&mutex);
      return 1;
   }
   sprintf(SharedContent, 
      "more lines\n%s\n%s\n%s\n%s\n%s\n%s\n%s",
      OverallParamValues[0],
      OverallParamValues[1],
      OverallParamValues[2],
      OverallParamValues[3],
      OverallParamValues[4],
      OverallParamValues[5],
      Melody
   );
   pthread_mutex_unlock(&mutex);
   return 0;
}


int UpdateMelody()
{
   int phrasecount;
   int i, maxlen=0, stop = 0, blockcount = 1, blocklen = 1024;
   int* lengths;
   WordListElement *currentword[10];
   char *CombinedPhrase = NULL;
   ClientListElement* currClient = ClientList;
   int beginning = 1;
   pthread_mutex_lock(&mutex);
   phrasecount = clientCount;
   
   lengths = (int*)malloc(phrasecount*sizeof(int));
   if(lengths==NULL)
   {
      printf("E013: Memory allocation failed\n");
      pthread_mutex_unlock(&mutex);
      return 1;
   }
   CombinedPhrase = (char*)malloc(blocklen);
   if(CombinedPhrase==NULL)
   {
      printf("E014: Memory allocation failed\n");
      free(lengths);
      pthread_mutex_unlock(&mutex);
      return 1;
   }
   CombinedPhrase[0] = '\0';

   for(i=0;i<phrasecount;i++)
   {
      wordlists[i]=split(currClient->Melody, &lengths[i]);
      currentword[i] = wordlists[i];
      if(lengths[i]>maxlen)
         maxlen = lengths[i];
      currentword[i] = wordlists[i];
      currClient = currClient->next;
   }
   do
   {
      for(i=0;i<phrasecount;i++)
      {
//+3 is for ending zero, blank and eventual new line if this is
//the last addition
         if(strlen(CombinedPhrase)+3+strlen(currentword[i]->word)
            >=blocklen*blockcount)
         {
            char* oldCombinedPhrase = CombinedPhrase;
            blockcount++;
            CombinedPhrase = (char*)malloc(blocklen*blockcount);
            if(CombinedPhrase==NULL)
            {
               printf("E015: Memory allocation failed\n");
               free(lengths);
               free(oldCombinedPhrase);
               pthread_mutex_unlock(&mutex);
               return 1;
            }
            strcpy(CombinedPhrase, oldCombinedPhrase);
            free(oldCombinedPhrase);
         }
         if(!beginning)
            strcat(CombinedPhrase, " ");
         else
            beginning = 0;
         strcat(CombinedPhrase, currentword[i]->word);
         currentword[i] = currentword[i]->next;
         if(currentword[i]==NULL)
            currentword[i] = wordlists[i];
      }
      stop++;
   }
   while(stop<maxlen);
   strcat(CombinedPhrase, "\n");
   WordListArrayCleanup(phrasecount);
   if(Melody != NULL)
      free(Melody);
   Melody = CombinedPhrase;  
   pthread_mutex_unlock(&mutex);
   return 0;
}


int findParam(char *p)
{
   for(int i=0;i<5;i++)
   {
      int n = strlen(OverallParams[i]);
      if(strncmp(p, OverallParams[i], n))
         return i;
   }
   return -1;
}

char * RemoveParamsFromInput(char *input)
{
   char *afterParams = NULL;
   char *paramsEnd = strrchr(input, '/');
   char *messageEnd;
   if(paramsEnd==NULL)
      return NULL;
   do
   { paramsEnd++; }
   while(*paramsEnd>='0'&&*paramsEnd<='9');
   messageEnd = strstr(paramsEnd, "end lines");
   if(messageEnd == NULL)
   {
      printf("E016: Incoming message missing end marker.\n");
      return NULL;
   }
   messageEnd--;
   *messageEnd = '\0';
   afterParams = (char *)malloc(strlen(paramsEnd)+1);
   if(afterParams==NULL)
   {
      printf("E017: Memory allocation failed\n");
      return NULL;
   }
   strcpy(afterParams, paramsEnd);
   return afterParams;
}

char* CommonParamsAndAllInputs(int index)
{
   int length;
   int j;
   char *response;
   ClientListElement* currClient = ClientList;
   length = 0;
   while(currClient!=NULL)
   {
      length+=(strlen(currClient->Name)+1);
      length+=2;
      if(currClient->index==index)
      {
         for(j=0;j<OverallParamCount;j++)
            length+=(strlen(OverallParams[j])
               +2+strlen(OverallParamValues[j])+1);
      }
      length+=strlen(currClient->Input0);
      length++;
      length+=strlen("end part\n");
      currClient = currClient->next;
   }
   length += strlen("end lines\n");
   response = (char *)malloc(length);
   if(response==NULL)
   {
      printf("E018: Memory allocation failed\n");
      return NULL;
   }
   sprintf(response, "%d\n", clientCount);
   currClient = ClientList;
   while(currClient!=NULL)
   {
      char currIndex[3];
      sprintf(currIndex, "%d\n", currClient->index);
      strcat(response, currClient->Name);
      strcat(response, currIndex);
      if(currClient->index==index)
      {
         for(j=0;j<OverallParamCount;j++)
         {
            strcat(response, OverallParams[j]);
            strcat(response, ": ");
            strcat(response, OverallParamValues[j]);
            if(j<OverallParamCount-1)
               strcat(response, "\n");
         }
      }
      strcat(response, currClient->Input0);
      strcat(response, "\nend part\n");
      currClient = currClient->next;
   }
   strcat(response, "end lines\n");
   return response;
}


#define MAXMESSAGE 5000

/*
 * This will handle connection for each client
 * */
void *connection_handler(void *client)
{
   int i;
   ClientListElement* myClient = (ClientListElement*)client;
   //Get the socket descriptor
   int sock = myClient->Socket;
   int read_size;
   char welcome_message[] = "0\nWelcome\n";
   char client_message[MAXMESSAGE];
   client_message[0]='\0';

   read_size = recv(sock , client_message , MAXMESSAGE, 0);
   if(read_size>0)
   {
      if(myClient->Name!=NULL)
         free(myClient->Name);
      myClient->Name = (char*)malloc(read_size+1);
      if(myClient->Name==NULL)
      {
         printf("E019: Memory allocation failed\n");
         return 0;
      }
      strcpy(myClient->Name, client_message);
      printf("New client: %s\n", myClient->Name);
      welcome_message[0] = myClient->index+'0';
      write(sock , welcome_message , strlen(welcome_message));
      while(1)
      {
         read_size = recv(sock , client_message , MAXMESSAGE, 0);
         if(read_size<=0)
            break;
         *(client_message+read_size)='\0';
         if(strncmp(client_message, "get", 3)==0)
         {
            if(!myClient->NewState||SharedContent==NULL)
               write(sock, "N\n", 2);
            else
            {
               char *getresponse = 
CommonParamsAndAllInputs(myClient->index);
               if(getresponse!=NULL)
               {
                  write(sock, SharedContent, strlen(SharedContent));
                  write(sock, getresponse, strlen(getresponse));
                  free(getresponse);
               }
               else
                  write(sock, "Memory error", 12);
               myClient->NewState = 0;
            }
         }
         else
         if(strncmp(client_message, "put", 3)==0)
         {
            int iRes;
            char *pstart, *pend, *input0;
            ClientListElement *currClient = ClientList;
            while(currClient!=NULL)
            {
               currClient->NewState = 1;
               currClient=currClient->next;
            }
            if(strlen(client_message)<5)
            {
               printf("Corrupted message. %s\n", client_message);
               continue;
            }
            pend = strchr(client_message, '\n');
            if(pend==NULL)
            {
               printf("Corrupted message. %s\n", client_message);
               continue;
            }
            pstart = pend+1;
            pend = strchr(pstart, '\n');
            if(pend==NULL)
            {
               printf("Corrupted message. %s\n", client_message);
               continue;
            }
            *pend = '\0';
            iRes = UpdateParams(pstart);
            pstart = pend+1;
            pend = strchr(pstart, '\n');
            if(pend==NULL)
            {
               printf("Corrupted message segment. %s\n", pstart);
               continue;
            }
            *pend = '\0';
            if(myClient->Melody!=NULL)
               free(myClient->Melody);
            myClient->Melody = (char*)malloc(strlen(pstart)+1);
            if(myClient->Melody==NULL)
            {
               printf("E020: Memory allocation failed\n");
               return 0;
            }
            strcpy(myClient->Melody, pstart);
            UpdateMelody();
            UpdateSharedContent();
            pstart = pend+1;
            input0 = RemoveParamsFromInput(pstart);
            if(input0!=NULL)
            {
               if(myClient->Input0!=NULL)
                  free(myClient->Input0);
               myClient->Input0 = input0;
            }
            write(sock, "Ok.\n", 4);
         }
         else
         if(strncmp(client_message, "bye", 3)==0)
            break;
         else
            write(sock, 
"Error! Type a command (get, put, bye, melody or param change)\n", 62);
      }
   }
   if(read_size>=0)
      write(sock, "bye\n", 4);
   removeClient(myClient->index);
   UpdateMelody();
   UpdateSharedContent();
   if(read_size <= 0)
   {
      printf("Client %d disconnected\n", myClient->index);
      fflush(stdout);
   }
      
   return 0;
}

void InitParams()
{
   int i;
   OverallParamValues = (char**)malloc(OverallParamCount*sizeof(char*));
   if(OverallParamValues==NULL)
   {
      printf("E021: Memory allocation failed\n");
      return;
   }
   for(i=0;i<OverallParamCount;i++)
   {
      OverallParamValues[i] = (char*)malloc(10);
      if(OverallParamValues[i]==NULL)
      {
         printf("E022: Memory allocation failed index=%d\n", i);
         return;
      }
   }
   strcpy(OverallParamValues[0], "150");
   strcpy(OverallParamValues[1], "128");
   strcpy(OverallParamValues[2], "400");
   strcpy(OverallParamValues[3], "4");
   strcpy(OverallParamValues[4], "6");
   strcpy(OverallParamValues[5], "1/1 1/2");
}

int LoadPortNumber()
{
   int retval;
   FILE *f = fopen("ServerConfiguration.txt", "rt");
   if(f==NULL)
   {
      printf("Could not open file ServerConfiguration.txt\n");
      return -1;
   }
   retval = fscanf(f, "%d", &PortNumber);
   if(retval!=1||PortNumber<=0)
   {
      printf("Invalid file ServerConfiguration.txt\n");
      return -1;
   }
   fclose(f);
   printf("Port number: %d\n", PortNumber);
   return 0;
}

int main(int argc , char *argv[])
{
   int socket_desc , new_socket , c;
   struct sockaddr_in server , client;
   char *message;

   InitParams(); 
   if(LoadPortNumber()!=0)
      return 1;
   printf("Server started.\n");
   //Create socket
   socket_desc = socket(AF_INET , SOCK_STREAM , 0);
   if (socket_desc == -1)
   {
      printf("Could not create socket");
   }
   
   //Prepare the sockaddr_in structure
   server.sin_family = AF_INET;
   server.sin_addr.s_addr = INADDR_ANY;
   server.sin_port = htons( PortNumber );
   
   //Bind
   if( bind(socket_desc,(struct sockaddr *)&server , sizeof(server))  < 0)
   {
      puts("bind failed");
      return 1;
   }
   
   //Listen
   listen(socket_desc , 3);
   
   //Accept and incoming connection
   puts("Waiting for incoming connections...");
   c = sizeof(struct sockaddr_in);
   while( (new_socket = accept(socket_desc, (struct sockaddr *)&client, (socklen_t*)&c)) )
   {
      if(clientCount>=9)
      {
         write(new_socket, "Too many clients. 9 is the maximum.\n", 36);
         continue;
      } 
      
      pthread_t sniffer_thread;  
      
      if(addClient(IdentifierCounter, new_socket)!=0)
         continue;

      IdentifierCounter++;
      UpdateMelody();
      if( pthread_create( &sniffer_thread , NULL ,  connection_handler , (void*) ClientListTail)  < 0)
      {
         perror("could not create thread");
         return 1;
      }
      
      printf("There are %d clients.\n", clientCount);      
   }
   
   if (new_socket <0)
   {
      perror("accept failed");
      return 1;
   }
   
   return 0;
}

